/************************************************************
 * $Id: objects.c 514 2006-03-06 16:44:41Z clerger $
 * Copyright (C) 1997, 2003, 2004, 2006 Eric de la Clergerie
 * ------------------------------------------------------------
 *
 *   DyALog objects --
 *
 * ------------------------------------------------------------
 * Description
 *
 * ------------------------------------------------------------
 */

//#define VERBOSE

#include "libdyalog.h"
#include "builtins.h"

#ifdef STAT

STAT_LIGHT_DEF(loop)		/* List of entries for loop stat in object.h */

#define V( entry )     STAT_USE(loop,entry)

static void stat_loop_nice_display() __attribute__ ((destructor)); 

void stat_loop_nice_display()
{
  printf(
	 
	 "--------------------------------------------------\n"
	  "STAT <execution>\n\n"
	  "Selections    : %4d [Std] %4d [Immediate] %5d [Last]\n"
	  "Applications run for %5d objects [%d direct + %d reverse]\n"
	  "    Delayed %6d and Tried %6d direct  applications\n"
	  "    Delayed %6d and Tried %6d reverse applications\n"
	  "    Tried %6d unifications whose %6d succeeded \n"
	  "Subsumptions run for %5d objects\n"
	  "    Tried %6d tests and kept %d objects\n"
	  "Creations     : %5d [= %5d items + %5d transitions + %5d answers ]\n"
	  "Tabulations   : %5d [= %5d items + %5d transitions + %5d answers ]\n"
	  "--------------------------------------------------\n" ,

				/* Line Selections */
	 V( sl_schedule_std ),
	 V( sl_schedule_first ),
	 V( sl_schedule_last ),

				/* Line Applications */
	 V( sl_direct_applications ) + V( sl_reverse_applications ),
	 V( sl_direct_applications ),
	 V( sl_reverse_applications ),
				/* Line Try Direct */
	 V( sl_tried_direct_applications ) - V( sl_tried_direct_applications2 ),
	 V( sl_tried_direct_applications2 ),
				/* Line Try Reverse */
	 V( sl_tried_reverse_applications ) - V( sl_tried_reverse_applications2 ),
	 V( sl_tried_reverse_applications2 ),
				/* Line Try Unification */
	 V( sl_unifications ),
	 V( sl_succeded_unifications ),
				/* Line Subsumptions */
	 V( sl_subsumptions ), 
	 V( sl_tried_subsumptions ),
	 V( sl_subsumptions ) - V( sl_succeded_subsumptions ),
				/* Line Creations */
	 V( sl_items ) + V( sl_transitions ) + V( sl_answers ),
	 V( sl_items ),
	 V( sl_transitions ),
	 V( sl_answers ),
				/* Line Installations */
	 V( sl_installed_items ) + V(sl_installed_transitions ) + V( sl_installed_answers ),
	 V( sl_installed_items ),
	 V( sl_installed_transitions ),
	 V( sl_installed_answers )

	 );
}

#undef V

#endif

extern tabobj_t agenda_next();
extern void tabulation_add( tabobj_t );

obj_t dyalog_solutions = BNIL;

/**********************************************************************
 * Constructors
 **********************************************************************/

void
save_object( tabseed_t seed )
{
    unsigned long k;
    obj_t env;
    Collapse_Unwrap(seed->model, R_TRANS_KEY, k, env);
    LVALUE_R_OBJECT = REG_VALUE(ALLOCATE_TABOBJ(seed,k,env,R_BACKPTR,R_LEVEL));
}

void
save_light_object( tabseed_t seed )
{
  LVALUE_R_OBJECT =  REG_VALUE(ALLOCATE_LIGHT_TABOJ(seed,R_BACKPTR,R_LEVEL));
}

tabobj_t
rt_install_database( SP(A) )
{
    fol_t type = find_folsmb("*DATABASE*",0);
    tabseed_t seed = ALLOCATE_TABSEED( -1, type, A, (obj_t) 0, 0);
    unsigned long  k;
    obj_t  env;
    tabobj_t obj;
    
    Collapse_Unwrap( A , Sk(A), k, env );
    obj = ALLOCATE_TABOBJ(seed,k,env,BINT( f_init ),0);
    tabulation_add( obj );
    return obj;
}

/**********************************************************************
 * Small Run Time functions
 **********************************************************************/

void
build_backptr( unsigned long type, fol_t trace )
{
    obj_t item = (obj_t) R_ITEM;
    switch (type) {
        case 0:                 /* INIT */
            LVALUE_R_BACKPTR = REG_VALUE(BINT( (unsigned long) f_init ));
            break;
        case 1:                 /* CALL */
            LVALUE_R_BACKPTR = REG_VALUE(BINT( (unsigned long) f_call ));
            break;
        case 3: /* JOIN */
            LVALUE_R_BACKPTR = REG_VALUE(MAKE_PAIR(BINT( (unsigned long) f_join ), MAKE_PAIR( (obj_t) R_TRANS, (obj_t) R_ITEM )));
            break;
        case 4: /* REVERSE AND */
            LVALUE_R_BACKPTR = REG_VALUE(MAKE_PAIR(BINT( (unsigned long) f_rand ), MAKE_PAIR((obj_t) R_TRANS, (obj_t) R_ITEM )));
//            R_BACKPTR = MAKE_PAIR(BINT( (unsigned long) f_and ), MAKE_PAIR((obj_t) R_ITEM, (obj_t) R_TRANS ));
//                        R_BACKPTR = MAKE_PAIR(BINT( (unsigned long) f_and ), MAKE_PAIR((obj_t) R_TRANS, (obj_t) R_ITEM ));
            break;
        case 10:             /* indirect cases */
        case 11:
        case 12:
        case 13:
        case 14:
            build_backptr( type-10, (fol_t)0);
            LVALUE_R_BACKPTR = REG_VALUE(MAKE_PAIR(BINT( (unsigned long) f_indirect ),MAKE_PAIR(R_BACKPTR,BNIL)));
            break;
        default: /* 2 AND */
//            V_LEVEL_DISPLAY(V_DYAM,"\tBACKPTR AND TRANS=%d
//            ITEM=%d\n",(int)R_TRANS,(int)R_ITEM);
            if (forest_indirect(R_TRANS->backptr))
                item = (obj_t)0;
            LVALUE_R_BACKPTR = REG_VALUE(MAKE_PAIR(BINT( (unsigned long) f_and ), MAKE_PAIR( (obj_t) R_TRANS, item )));
    }
    if (trace && trace != FOLNIL) {
        LVALUE_R_BACKPTR= REG_VALUE(MAKE_PAIR( BINT( (unsigned long) f_trace),
                                               MAKE_PAIR( (obj_t) trace, R_BACKPTR )));
    }
}

/**********************************************************************
 * RT Treatement functions
 **********************************************************************/

extern void DyALog_Fail();

static void
treat_selected( tabobj_t o )
{
  fkey_t k = Trail_Object( o );
  fun_t code = (fun_t) o->seed->application_code;
  tabobj_t saved_trans = R_TRANS;
  fkey_t saved_trans_k = R_TRANS_KEY;

  LVALUE_R_CP= REG_VALUE(DyALog_Fail);
  
  STAT_UPDATE(loop, sl_selections );
  V_LEVEL_DISPLAY( V_DYAM, "--------------------------------------------------\n"
	     "%&e\n%&l\n" 
	     "Selected: [%&k] %&s\n", k, o->seed->model, k );

//  V_LEVEL_DISPLAY(V_DYAM,"\tR_ITEM=%&f\n",(fol_t) R_ITEM);
    
  o->examined = 2;		/* 2 is code for 'under selection' */

  if (code) {
      R_LOAD_TRANS(o,k);
      Dyam_DyALog(code);
      R_LOAD_TRANS(saved_trans, saved_trans_k);
  }
  
  o->examined = 1;		/* 1 is code for 'treated' */

  Untrail_Object;

  V_LEVEL_DISPLAY( V_DYAM, "Unloading Selected [%&e]\n" );

}

Bool
treat_item( tabobj_t item,
              long id,
              fol_t type,
              fun_t code,
              tabobj_t transition,
              fkey_t transition_key
              )
{
        void **tab = &item->seed->application_code+1;
        fol_t entry = ((item_comp_t) tab + id)->item_comp;

        STAT_UPDATE(loop, sl_tried_direct_applications );
        
        if ( item->examined && ( type == item->seed->type )   &&  entry ) {
            fun_t init_code = ((item_comp_t) tab + id)->init_code;
            fkey_t item_key = LOAD_OBJECT( item );
    
            STAT_UPDATE(loop, sl_tried_direct_applications2 );
            V_LEVEL_DISPLAY( V_DYAM, "     *try with %&s\n", item->seed->model, item_key );

            R_LOAD_ITEM( item, item_key, entry );
            R_LOAD_TRANS( transition, transition_key );
            LVALUE_R_P =  REG_VALUE((fun_t) Adjust_Address(code));
            if ((init_code))
                LVALUE_X(0) = REG_VALUE((fun_t) Adjust_Address(init_code));
            Succeed;
        } else
            Fail;
}

Bool
treat_transition( tabobj_t transition, long id, fol_t type)
{
    void **tab = &transition->seed->application_code+1;
    fun_t entry = ((trans_comp_t) tab + id)->code;
    
    STAT_UPDATE(loop, sl_tried_reverse_applications );
    
    if ( transition->examined  && ( type == transition->seed->type ) && entry ) {
        fkey_t transition_key = LOAD_OBJECT( transition );
        fun_t init_code = ((trans_comp_t) tab + id)->init_code;
        
        STAT_UPDATE(loop, sl_tried_reverse_applications2 );
        V_LEVEL_DISPLAY( V_DYAM, "     *try with %&s\n",
                         transition->seed->model, transition_key );
                
        R_LOAD_TRANS( transition, transition_key );
            
        LVALUE_R_P = REG_VALUE((fun_t) Adjust_Address(entry));            
        if ((init_code))
            LVALUE_X(0) = REG_VALUE((fun_t) Adjust_Address(init_code));
        
        Succeed;
        
    } else
        Fail;
}

#define B_TYPE( b )  (backptr_t) (PAIRP( b ) ? CINT( CAR( b )) : CINT( b ))
#define B_CAR( b )   (CAR( CDR( b )))
#define B_CDR( b )   (CDR( CDR( b )))

Bool
treat_generalizer( tabobj_t generalizer, fol_t model, fkey_t k_model )
{
  Bool v = 0;
  fkey_t generalizer_key = Trail_Object( generalizer );

  STAT_UPDATE(loop, sl_tried_subsumptions );
  V_LEVEL_DISPLAY( V_DYAM, "     *try generalizer %&s ",
                   generalizer->seed->model, generalizer_key);

  if ( sfol_subsume( generalizer->seed->model, generalizer_key,
		     model, k_model ) ) {

    STAT_UPDATE(loop, sl_succeded_subsumptions );
    V_LEVEL_DISPLAY( V_DYAM, "( *** FOUND *** )\n" );
    
    switch (B_TYPE(R_BACKPTR)) {
        case f_init:
        case f_call:

                /* do nothing */
            
            break;

        case f_indirect:        /* assume that gen->backptr is also indirect */
            generalizer->backptr =
                MAKE_PAIR( BINT((unsigned long) f_indirect),
                           MAKE_PAIR( MAKE_PAIR( BINT( (unsigned long) f_or ),
                                                 MAKE_PAIR(
                                                     B_CAR(R_BACKPTR),
                                                     B_CAR(generalizer->backptr ))),
                                      BNIL));
            break;
            
        case f_trace:
            if (B_TYPE( generalizer->backptr ) == f_trace &&
                B_CAR(R_BACKPTR) == B_CAR( generalizer->backptr )) {
                generalizer->backptr =
                    MAKE_PAIR( BINT( (unsigned long) f_trace ),
                               MAKE_PAIR( B_CAR( R_BACKPTR),
                                          MAKE_PAIR( BINT( (unsigned long) f_or ),
                                                     MAKE_PAIR(
                                                         B_CDR(R_BACKPTR),
                                                         B_CDR(generalizer->backptr )))));
                break;
            }
                /* otherwise use default case */
            
        default:
            generalizer->backptr =
                MAKE_PAIR( BINT( (unsigned long) f_or ),
                           MAKE_PAIR( R_BACKPTR, generalizer->backptr ));
            break;
    }
    
    if (generalizer->level < R_LEVEL)
      generalizer->level = R_LEVEL;

    v = 1;
  } else {
      V_LEVEL_DISPLAY( V_DYAM, "\n" );
  }
  
  Untrail_Object;
  return v;
}

void
treat_answer()
{
  fol_t tuple  = FOLCMP_REF(R_OBJECT->seed->model,1);

  dyalog_solutions = MAKE_PAIR( (obj_t) R_OBJECT, dyalog_solutions );

  V_LEVEL_DISPLAY( V_DYAM, "----------------------------------------------------------------------\n" );

  dyalog_printf( "Answer:" );
  subst_display( tuple, R_TRANS_KEY );

  V_LEVEL_DISPLAY( V_DYAM, "----------------------------------------------------------------------\n" );
  
}

/**********************************************************************
 * Main DyALog Loop
 **********************************************************************/

void
dyalog_loop()
{
  tabobj_t o;
  V_LEVEL_DISPLAY( V_DYAM, "\n"
	     "--------------------------------------------------\n"
	     "Running ...\n"
	     "--------------------------------------------------\n\n"
	     );

  while ((o=agenda_next()))
    treat_selected( o );

}

/**********************************************************************
 * Dyam Instructions
 **********************************************************************/

void                            /* Dyam Instruction */
Dyam_Backptr(long Kind)
{
    V_LEVEL_DISPLAY( V_DYAM, "\tbuild backptr %d\n",Kind);
    build_backptr(Kind, (fol_t) 0);
}

void                            /* Dyam Instruction */
Dyam_Backptr_Trace(long Kind, fol_t Trace)
{
    fkey_t Sk(Trace)=R_TRANS_KEY;
    V_LEVEL_DISPLAY( V_DYAM, "\tbuild backptr with trace %d %&s\n",Kind,Trace,Sk(Trace));
    Deref(Trace);
    build_backptr(Kind,Trace);
}

void                            /* Dyam Instruction */
Dyam_Answer()
{
    V_LEVEL_DISPLAY( V_DYAM, "\tanswer\n");
    treat_answer();
}

void                            /* Dyam Instruction */
Dyam_Object( tabseed_t seed )
{
    V_LEVEL_DISPLAY( V_DYAM,"\tobject %&s\n",seed->model,R_TRANS_KEY);
    save_object(seed);
}

void                            /* Dyam Instruction */
Dyam_Light_Object( tabseed_t seed )
{
    V_LEVEL_DISPLAY( V_DYAM,"\tlight object %&s\n",seed->model,R_TRANS_KEY);
    save_light_object(seed);
}







